package rocq-runtime

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Tac2syn.SyntaxSource

Sourcetype 'a t

Type of notation syntax parsing 'a. Unlike Procq.Symbol.t it fully supports comparison and is marshallable.

Sourcetype 'a seq

Sequence of t.

Sourcetype 'a entry

Marshal-stable proxy for Procq.Entry.t.

Sourceval register_entry : ?name:string -> 'a Procq.Entry.t -> 'a entry

Must be called at toplevel, with non backtrackable entry. name defaults to the entry name but can be given another value if there is a conflict. Registering the same entry twice produces different entry values.

Pre-registered entries.

Constructors for t, copying Procq.Symbol constructors.

Sourceval nterm : 'a entry -> 'a t
Sourceval nterml : 'a entry -> string -> 'a t
Sourceval list0 : ?sep:string -> 'a t -> 'a list t
Sourceval list1 : ?sep:string -> 'a t -> 'a list t
Sourceval opt : 'a t -> 'a option t
Sourceval token : 'a Tok.p -> 'a t
Sourceval tokens : Procq.ty_pattern list -> unit t
Sourceval seq : 'a seq -> 'a t

Instead of rules we have the less general seq.

Sourceval nil : unit seq
Sourceval snoc : 'a seq -> 'b t -> ('a * 'b) seq