package rocq-runtime

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module WriteOnceArraySource

Sourcetype 'a t

Like a Parray, but the old pointers are invalidated instead of updated

Sourceval make : int -> 'a t
Sourceval get : int -> 'a t -> 'a option
Sourceval is_filled : int -> 'a t -> bool
Sourceval add : int -> 'a -> 'a t -> 'a t
Sourceval fill_remaining : 'a -> 'a t -> 'a t
Sourceval to_array : 'a t -> 'a array
Sourceval to_array_opt : 'a t -> 'a array option

The NoDupArray is still invalidated if the result is None

Sourcemodule Internal : sig ... end