package rocq-runtime

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module EvarnamesSource

Sourceval generate_goal_names : unit -> bool

Whether the Generate Goal Names option is enabled.

Sourcetype t

Represents an evar name map.

Sourceval empty : t

Returns an empty name map.

Sourceval add : Names.Id.t -> Evar.t -> ?parent:Evar.t -> t -> t

Adds a binding for the given undefined evar to the given basename. The absolute path is obtained using the parent of the evar.

Sourceval add_fresh : Names.Id.t -> Evar.t -> ?parent:Evar.t -> t -> t

Adds a (potentially fresh) binding for the given undefined evar to the given basename by first checking for conflicts.

Sourceval remove : Evar.t -> t -> t

Removes the name of the given evar. This indicates that the evar was defined, and therefore is no longer accessible by name.

Sourceval transfer_name : Evar.t -> Evar.t -> t -> t

Transfers the name of the first evar to the second.

Sourceval name_of : Evar.t -> t -> Libnames.full_path option

Returns the qualified name associated to the evar, if any.

Sourceval has_name : Evar.t -> t -> bool

Returns true if the evar has a name. Equivalent to name_of ev <> None but faster since it does not compute the fully qualified name of ev.

Sourceval has_unambiguous_name : Evar.t -> t -> bool

Returns true if the evar has a name that is unambiguous.

Sourceval resolve : Libnames.qualid -> t -> Evar.t

Resolves the given (partially) qualified name to an evar. If the name resolution failed, raises Not_found.