package rocq-runtime

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module UVars.AbstractContextSource

Sourcetype t

An abstract context serves to quantify over a graph of universes represented using de Bruijn indices, as in: u0, ..., u(n-1), Var i < Var j, .., Var k <= Var l |- term(Var 0 .. Var (n-1)) \-------------/ \-------------------------------/ \---------------------/ names for constraints expressed on de Bruijn judgement in abstract printing representation of the n univ vars context expected to use de Bruijn indices

Build an abstract context. PConstraints may be between sorts/universe variables.

Sourceval repr : t -> UContext.t

repr ctx is (Var(0), ... Var(n-1) |= cstr where n is the length of the context and cstr the abstracted UnivConstraints.t.

Sourceval hcons : t -> int * t
Sourceval empty : t
Sourceval is_empty : t -> bool
Sourceval is_constant : t -> bool

Empty instance, but may have constraints

Sourceval size : t -> int * int
Sourceval union : t -> t -> t

The constraints are expected to be relative to the concatenated set of universes

Sourceval instantiate : Instance.t -> t -> PConstraints.t

Generate the set of instantiated PConstraints.t *

Sourceval names : t -> bound_names

Return the names of the bound universe variables

Sourceval pr : (Sorts.QVar.t -> Pp.t) -> (Univ.Level.t -> Pp.t) -> ?variance:Variance.t array -> t -> Pp.t