package rocq-runtime

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Hopcroft.MakeSource

Parameters

Signature

Sourcetype label = Label.t
Sourcetype state = int
Sourcetype transition = {
  1. src : state;
  2. lbl : label;
  3. dst : state;
}
Sourcetype automaton = {
  1. states : int;
    (*

    The number of states of the automaton.

    *)
  2. partitions : state list list;
    (*

    A set of state partitions initially known to be observationally distinct. For instance, if the automaton has the list l as accepting states, one can set partitions = [l].

    *)
  3. transitions : transition list;
    (*

    The transitions of the automaton without duplicates.

    *)
}
Sourceval reduce : automaton -> state list array

Associate the array of equivalence classes of the states of an automaton