package rocq-runtime

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type Goptions.TableArgSource

The functor MakeTable declares a table. Vernacular Add, Remove, and Test commands pass each argument as a table_value; encode maps such values to stored elements, possibly globalizing qualified identifiers with the supplied environment. Tables that do not support one of the constructors should raise a user error from encode.

Sourcetype t
Sourcemodule Set : ArgSet with type elt = t
Sourceval encode : Environ.env -> table_value -> t
Sourceval subst : Mod_subst.substitution -> t -> t
Sourceval check_local : Libobject.locality -> t -> unit
Sourceval discharge : t -> t

Elements which cannot be discharged should only be added with Local

Sourceval printer : t -> Pp.t
Sourceval title : string
Sourceval member_message : t -> bool -> Pp.t