package rocq-runtime

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module PrintingFlags.ExternSource

Sourcemodule FactorizeEqns : sig ... end
Sourcemodule Records : sig ... end
Sourcetype t = {
  1. use_implicit_types : bool;
  2. records : Records.t;
  3. implicits : bool;
    (*

    When implicits is on then implicits_explicit_args tells how implicit args are printed. If on, implicit args are printed with the form (id:=arg) otherwise arguments are printed normally and the function is prefixed by "@".

    *)
  4. implicits_explicit_args : bool;
  5. implicits_defensive : bool;
  6. coercions : bool;
  7. parentheses : bool;
  8. notations : bool;
  9. raw_literals : bool;
  10. projections : bool;
  11. float : bool;
  12. factorize_eqns : FactorizeEqns.t;
  13. depth : int option;
}
Sourceval make_raw : t -> t
Sourceval current : unit -> t
Sourceval current_ignore_raw : unit -> t